//****************************************************************
//*                                                              *
//*    Instr.h : Header file for the object class hierarchy for  *
//*              the instructions.                               *
//*                                                              *
//* (c) LITA, university of PAU (France), summer 1993.           *
//****************************************************************

#include <String.h>
#include <iostream.h>
#include "Objects.h"
#include "Expr.h"
#include "Instr.h"

//***************************************
//**                                   **
//**  Objects decribing an instruction **
//**  inside an abstract tree.         **
//**                                   **
//***************************************

/*
                        Instruction

             Single_Instruction                Complex_Instruction


                           Single_Instruction

*:=      *job        io       signal        control      object

        Attach      Put       Raise         Return       Wind  
        Resume      Get                     Call         Inner
        Stop        Read                    Exit         Kill
        Detach      ReadLn                               Generator
        Terminate   Write
                    WriteLn


                           Complex_instructions

   *Loop                         Block                  Condition

   *For                        Pref Block              *If
   *While                                               Case
*/

virtual void Instruction::Print( ostream& stream )
{
  stream << " \"" << kind << "\" ";
}

ostream& operator << ( ostream& stream , Instruction TheInstr )
{
  TheInstr.Print( stream );
}

OneArgInstr::OneArgInstr( Expression * TheExpr, String TheStr, Location *TheLoc ) :
             Instruction( TheStr, TheLoc )
{
  argument = TheExpr;
}

void OneArgInstr::Print( ostream& stream )
{
  Instruction::Print( stream );
  stream << "(" << *argument << ");\n";
}

TwoArgInstr::TwoArgInstr( Expression *TheExpr1, Expression *TheExpr2,
                          String TheStr , Location *TheLoc) :
             Instruction( TheStr, TheLoc )
{
  arg1 = TheExpr1;
  arg2 = TheExpr2;
}

void TwoArgInstr::Print( ostream& stream )
{
  Instruction::Print( stream );
  stream << "(" << *arg1 << "," << *arg2 << ");";
}

Affectation::Affectation( Expression *L_Value, Expression *R_Value ) :
             TwoArgInstr( L_Value , R_Value, String("Affectation"),
                          new Location(*L_Value->place + *R_Value->place) )
{
  switch( L_Value->returntype )
  {
    case Ident:
      /* Here we should verify if the returntype of the Expression is coherent
         with the type of the Identifier ( if it is declared ). */
      break;
    default:
      L_Value->Print( cout );
      cout << " is not an Identifier .\n" ;
      break;
  }
}

void Affectation::Print( ostream& stream )
{
  arg1->Print( stream );
  cout << " := ";
  arg2->Print( stream );
}

Attach::Attach( Expression *TheExpr, Location *TheLoc ) :
        OneArgInstr( TheExpr, String("Attach"),
                     new Location( *TheLoc + *TheExpr->place ))
{
  switch( TheExpr->returntype )
  {
     case Ident:
     case ObjExpr:
     case ObjConst:
       /* Here we should verify if the Expression
          is coherent with a coroutine object. */
       break;

     case ExprError:
       /* Here we test if the expression is correct or not.
          If it is not correct, a report of the error has already been
          emited and we have to go ahead without any comment for avoiding
          errors provoked by the wrong expression. */
       break;

     default:
       /* Here the Expression is correct according to the rules of building
          an expression but the expression is not an object type so we emit
          an error saying that there is a mismatched expression used with
          the Attach instruction. */
       cout << " Operand \"";
       TheExpr->Print( cout );
       cout << "\" for instruction Attach is wrong.\n";
       break;
   }
}

Resume::Resume( Expression *TheExpr, Location *TheLoc ) :
        OneArgInstr( TheExpr, String("Resume"),
                     new Location( *TheLoc + *TheExpr->place ) )
{
  switch( TheExpr->returntype )
  {
     case Ident:
     case ObjExpr:
     case ObjConst:
       /* Here we should verify if the Expression
          is coherent with a process object. */
       break;

     case ExprError:
       /* Here we test if the expression is correct or not.
          If it is not correct, a report of the error has already been
          emited and we have to go ahead without any comment for avoiding
          errors provoked by the wrong expression. */
       break;

     default:
       /* Here the Expression is correct according to the rules of building
          an expression but the expression is not an object type so we emit
          an error saying that there is a mismatched expression used with
          the Resume instruction. */
       cout << " Operand \"";
       TheExpr->Print( cout );
       cout << "\" for instruction Resume is wrong.\n";
       break;
   }
}

Stop::Stop( Expression *TheExpr, Location *TheLoc ) :
      OneArgInstr( TheExpr, String("Stop"),
                   new Location( *TheLoc + *TheExpr->place) )
{
  switch( TheExpr->returntype )
  {
     case Ident:
     case ObjExpr:
     case ObjConst:
       /* Here we should verify if the Expression
          is coherent with a process object. */
       break;

     case ExprError:
       /* Here we test if the expression is correct or not.
          If it is not correct, a report of the error has already been
          emited and we have to go ahead without any comment for avoiding
          errors provoked by the wrong expression. */
       break;

     default:
       /* Here the Expression is correct according to the rules of building
          an expression but the expression is not an object type so we emit
          an error saying that there is a mismatched expression used with
          the instruction Stop. */
       cout << " Operand \"";
       TheExpr->Print( cout );
       cout << "\" for instruction Stop is wrong.\n";
       break;
   }
}

Stop::Stop( Location *TheLoc ) : OneArgInstr( NULL, String("Stop"), TheLoc ) {}

void Stop::Print( ostream& stream )
{
  if (argument)
    OneArgInstr::Print( stream );
  else
    cout << "Stop;\n";
}

Get::Get( Expression *TheArg, Location *TheLoc) :
  OneArgInstr( TheArg, String("Get"),
               new Location(*TheArg->place + *TheLoc) )
{
  switch(TheArg->returntype)
  {
    case Ident:
    case ExprError:
      break;

    default:
      cout << " Operand \"";
      TheArg->Print( cout );
      cout << "\" for instruction Get is wrong.\n";
      break;
  }
}

Put::Put( Expression *TheArg, Location *TheLoc) :
     OneArgInstr( TheArg, String("Put"),
                  new Location(*TheArg->place + *TheLoc) )
{
  switch(TheArg->returntype)
  {
    case Ident:
    case ExprError:
      break;

    default:
      cout << " Operand \"";
      TheArg->Print( cout );
      cout << "\" for instruction Put is wrong.\n";
      break;
  }
}

Block::Block( Instruction *FirstInstr, Location *TheLoc ) :
       Instruction( String("Block"), TheLoc ), InstrList( FirstInstr )
{

// We have to test the pointer because a block of instruction could be empty.
// i.e. : for i := 1 to 2 do od;
//                        \___/-> empty block;
  if (FirstInstr)
    *place = *place + *FirstInstr->place;

}

Block& operator += ( Block& TheBlock, Instruction *TheInstr )
{
  TheBlock.InstrList += TheInstr;
// The test is not worth here because the block is expanded only if there is
// a new instruction who of course is not empty.
  *TheBlock.place = *TheBlock.place + *TheInstr->place;
  return TheBlock;
}

void Block::Print( ostream& stream )
{
  stream << "Block\n";
  InstrList.Print( stream );
  stream << "End of Block.\n";
}

ListOfInstr::ListOfInstr( Instruction *TheInstr )
{
  Instruct = TheInstr;
  NextInst = NULL;
}

ListOfInstr& operator += ( ListOfInstr& TheList , Instruction *TheInstr )
{
  ListOfInstr *Current = &TheList;

  while (Current->NextInst != NULL ) Current = Current->NextInst;
  Current->NextInst = new ListOfInstr( TheInstr );
  return TheList;
}

void ListOfInstr::Print( ostream& stream )
{
  ListOfInstr *Current;

  Current = NextInst;
  Instruct->Print( stream );
  stream << "\n";

  while (Current != NULL)
  {
    Current->Instruct->Print( stream );
    stream << "\n";
    Current = Current->NextInst;
  }

}

ConditionIf::ConditionIf( Expression *TheCondition,
                          Block *Block1, Block *Block2, Location *TheLoc ) :
             Instruction( String( " If " ),
  new Location( *TheLoc + *Block1->place + *Block2->place ) ),
             Condition( TheCondition ), BlockThen( Block1 ), BlockElse( Block2 )
{
  switch( TheCondition->returntype )
  {
    case BoolExpr:
    case BoolConst:
    case Ident:
    case ExprError:
      break;

    default:
      cout << " Bad Expression \"";
      TheCondition->Print( cout );
      cout << " for If condition.\n";
      break;
  }
}

void ConditionIf::Print( ostream& stream )
{
  cout << "If (";
  Condition->Print( cout );
  cout << ")\nThen \n";
  BlockThen->Print( cout );
  if (BlockElse)
  {
    cout <<"Else\n";
    BlockElse->Print( cout );
  }
}

void Loop::Print( ostream& stream )
{
  stream << " DO\n";
  Body->Print( stream );
  stream << "\nOD\n";
}

While::While( Expression *TheExpr, Block *TheBlock, Location *TheLoc ) :
       Loop( TheBlock, String("While") ), Condition( TheExpr )
{
  if (TheExpr)
  {
    *place = *TheLoc + *TheBlock->place;
    switch( TheExpr->returntype )
    {
      case Ident:
      case BoolConst:
      case BoolExpr:
      case ExprError:
        break;

      default:
        cout << "Wrong expression \"";
        TheExpr->Print( cout );
        cout << "\" for Instruction While.\n";
        break;
    }
  }
}

void While::Print( ostream& stream )
{
  stream << "While ";
  Condition->Print( stream );
  Loop::Print( stream );
}

For::For( Expression *TheVar, Expression *Initial, Expression *Final,
          Expression *Step,   Block *TheBlock, Location *TheLoc ) :
     Loop( TheBlock, String("For") ), Counter( TheVar ),
     CounterInit( Initial ), CounterStop( Final ), CounterStep( Step )
{
  *place = *TheLoc + *TheVar->place + *Initial->place + *Final->place +
           *Step->place + *TheBlock->place;
  switch( CounterInit->returntype )
  {
    case Ident:
    case IntegerConst:
    case IntegerExpr:
    case RealExpr:
    case RealConst:
    case ExprError:
      break;

    default:
      cout << " Bad Expression type for Initial value \"";
      CounterInit->Print( cout );
      cout << "\" of variable ";
      Counter->Print( cout );
      cout << ".\n";
  }

  switch( CounterStop->returntype )
  {
    case Ident:
    case IntegerConst:
    case IntegerExpr:
    case RealExpr:
    case RealConst:
    case ExprError:
      break;

    default:
      cout << " Bad Expression type for Initial value \"";
      CounterInit->Print( cout );
      cout << "\" of variable ";
      Counter->Print( cout );
      cout << ".\n";
  }

  switch( CounterStep->returntype )
  {
    case Ident:
    case IntegerConst:
    case IntegerExpr:
    case RealExpr:
    case RealConst:
    case ExprError:
      break;

    default:
      cout << " Bad Expression type for Step value \"";
      CounterStep->Print( cout );
      cout << ".\n";
  }

}

void For::Print( ostream& stream )
{
  stream << "For ";
  Counter->Print( cout );
  stream << " := ";
  CounterInit->Print( cout );
  stream << " TO ";
  CounterStop->Print( cout );
  stream << " Step ";
  CounterStep->Print( cout );
  Loop::Print( stream );
}
