//****************************************************************
//*                                                              *
//*    Instr.h : Header file for the object class hierarchy for  *
//*              the instructions.                               *
//*                                                              *
//* (c) LITA, university of PAU (France), summer 1993.           *
//****************************************************************

//***************************************
//**                                   **
//**  Objects decribing an instruction **
//**  inside an abstract tree.         **
//**                                   **
//***************************************

/*
                        Instruction

             Single_Instruction                Complex_Instruction


                           Single_Instruction

 :=       job        io       signal        control      object

        Attach      Put       Raise         Return       Wind  
        Resume      Get                     Call         Inner
        Stop        Read                    Exit         Kill
        Detach      ReadLn                               Generator
        Terminate   Write
                    WriteLn


                           Complex_instructions

   Loop                         Block                  Condition

   For                        Pref Block               If
   While                                               Case
*/

class Instruction
{
  public:
  String kind;
  Location *place;

  Instruction( String TheStr, Location *TheLoc ) :
  kind( TheStr ), place( TheLoc )
  {}

  virtual void Print( ostream& );

  friend ostream& operator << ( ostream& , Instruction );
} ;

class OneArgInstr : public Instruction
{
  public:
  Expression *argument;

  OneArgInstr( Expression * , String, Location * );
  virtual void Print( ostream& );
} ;

class TwoArgInstr : public Instruction
{
  public:
  Expression *arg1,*arg2;

  TwoArgInstr( Expression *, Expression *, String, Location *);
  virtual void Print( ostream& );
} ;

class Affectation : public TwoArgInstr
{
  public:
  Affectation( Expression *, Expression * );
  void Print( ostream& );
} ;

typedef enum
{
  AttachInstr,
  ResumeInstr,
  StopInstr
} OneArgJob;

class Attach : public OneArgInstr
{
  public:
  Attach( Expression *, Location * );
} ;

class Resume : public OneArgInstr
{
  public:
  Resume( Expression *, Location * );
} ;

class Detach : public Instruction
{
  public:
  Detach( Location *TheLoc ) : Instruction( String( "Detach" ), TheLoc ) {}
} ;

class Terminate : public Instruction
{
  public:
  Terminate( Location *TheLoc ) : Instruction( String( "Terminate" ), TheLoc ) {}
} ;

class Stop : public OneArgInstr
{
  public:
  Stop( Expression *, Location * );
  Stop( Location * );
  void Print( ostream& );
} ;

//**
//** These class describes the I/O instructions of LOGLAN
//**
//** Each Instruction taking a variable argument number
//** for example "writeln( output,a,b )"
//** is simplified into the following Instructions :
//**   write( output, a );
//**   write( output, b );
//**   writeln( output );
//**

class Read : public OneArgInstr
{
  public:
  Read( Expression *, Location * );
} ;

class Readln : public Instruction
{
  public:
  Readln( Location * );
} ;

class Write : public OneArgInstr
{
  public:
  Write( Expression *, Location * );
} ;

class Writeln : public Instruction
{
  public:
  Writeln( Location * );
} ;

class Get : public OneArgInstr
{
  public:
  Expression *TheFile, *TheArg;
  Get( Expression *, Location * );
} ;

class Put : public OneArgInstr
{
  public:
  Expression *TheFile, *TheArg;
  Put( Expression *, Location * );
} ;

//**
//** This is the only Instruction concerning signals.
//**

class Raise : public OneArgInstr
{
  public:
  Raise( Expression *TheExpr, Location *TheLoc ) :
       OneArgInstr( TheExpr, String("Raise"), TheLoc ) {}
} ;

//**
//**  Class container for a set of Instructions. We call it a block.
//**  Tool class for containing several instructions and manipulate them.
//**

class ListOfInstr
{
  Instruction *Instruct;
  ListOfInstr *NextInst;
  public:
  friend ListOfInstr& operator += ( ListOfInstr& , Instruction * );
  ListOfInstr( Instruction *TheInstr );
  void Print( ostream& );
} ;

class Block : public Instruction
{
  ListOfInstr InstrList;
  public:
  Block( Instruction *FirstInstr, Location *TheLoc );
  friend Block& operator += ( Block& TheBlock, Instruction *TheInstr );
  void Print( ostream& );
} ;

//**
//** Class conditionnal If : A class with 2 Blocks conditionnaly executed.
//**

class ConditionIf : public Instruction
{
  Block      *BlockThen, *BlockElse;
  Expression *Condition;
  public:
  ConditionIf( Expression *, Block *, Block *, Location * );
  void Print( ostream & );
} ;

//**
//** Class Loop : A class for hanling loop construction
//**

class Loop : public Instruction
{
  Block *Body;
  public:
  Loop( Block *TheBlock, String TheName ) :
      Instruction( TheName, TheBlock->place ), Body( TheBlock ) {};
  void Print( ostream& );
} ;

class While : public Loop
{
  Expression *Condition;
  public:
  While( Expression *, Block *, Location * );
  void Print( ostream& );
} ;

class For : public Loop
{
  Expression *Counter;
  Expression *CounterInit;
  Expression *CounterStop;
  Expression *CounterStep;
  public:
  For( Expression *,Expression *, Expression *, Expression *, Block *,Location * );
  void Print( ostream& );
} ;
