#include <stdlib.h>
#include <String.h>
#include "Objects.h"

#define LINE_WIDTH 256 // set the line width to anything you want.
                       // The Only assumption is that a column number
                       // is lesser than LINE_WIDTH

#define POSITION(line,column) ( (line) * LINE_WIDTH + (column) )

//**************************************
//* Class Location :                   *
//*  Class used for storing location   *
//*  of an object		       *
//**************************************

// Constructor with two arguments : the markers are set equal each other
// Start mark = End mark.

Location::Location( void )
{
  LineStart   = LineEnd   = 1;
  ColumnStart = ColumnEnd = 1;
}

Location::Location( int Line, int Column = 1 )
{
  if (Line < 0 || Column < 0 || Column > LINE_WIDTH)
    error("Incoherent values for initialisation of Location object.");
  LineStart   = LineEnd   = Line;
  ColumnStart = ColumnEnd = Column;
}

// Constructor with four arguments : the two first are for Start mark
// and the two last are for the End mark.

Location::Location( int First_Line, int First_Column,
	       int Last_Line, int Last_Column )
{
  if ((First_Line <= 0) || (First_Column <= 0) ||
      (Last_Line  <= 0) || (Last_Column  <= 0) ||
      (First_Column > LINE_WIDTH) || (Last_Column > LINE_WIDTH))
    error("Incoherent values for initialisation of Location object.");

  if ( POSITION(First_Line,First_Column) < POSITION(Last_Line,Last_Column) )
  {
    LineStart   = First_Line;
    LineEnd     = Last_Line;
    ColumnStart = First_Column;
    ColumnEnd   = Last_Column;
  }
  else
  {
    LineStart   = Last_Line;
    LineEnd     = First_Line;
    ColumnStart = Last_Column;
    ColumnEnd   = First_Column;
  }
}

Location::Location( const Location& TheLoc )
{
  LineStart   = TheLoc.LineStart;
  LineEnd     = TheLoc.LineEnd;
  ColumnStart = TheLoc.ColumnStart;
  ColumnEnd   = TheLoc.ColumnEnd;
}

// SetEnd method : Set the End mark to line number Line
//                 and column number Column.

void Location::SetEnd( int Line, int Column )
{

  if ( Line <= 0 || Column <= 0 || Column > LINE_WIDTH)
    error("Incoherent values for initialisation of Location object.");

  if ( POSITION(Line,Column) > POSITION(LineStart,ColumnStart) )
// The new End Mark is after the Start Mark so we just adjust the End Mark.
  {
    LineEnd     = Line;
    ColumnEnd   = Column;
  }
  else
  {
// The new End Mark is before the Start Mark so we have two swap the two marks.
    LineEnd     = LineStart;
    ColumnEnd   = ColumnStart;
    LineStart   = Line;
    ColumnStart = Column;
  }
}

// This function sets the beginning of the Location Object at
// the line Line and the column Column

void Location::Move( int Line , int Column )
{
  LineStart   = Line;
  ColumnStart = Column;
  if ( POSITION(LineStart,ColumnStart) > POSITION(LineEnd,ColumnEnd) )
  {
    ColumnEnd = ColumnStart;
    LineEnd   = LineStart;
  }
}

// This function simply advance the start position by 'distance' columns.
void Location::Tab( int distance )
{
  ColumnStart += distance;
  if (ColumnStart > LINE_WIDTH)
  {
    LineStart++;
    ColumnStart = ((ColumnStart - 1) % LINE_WIDTH) + 1;
  }
  if (POSITION(LineStart,ColumnStart) > POSITION(LineEnd,ColumnEnd) )
  {
    LineEnd   = LineStart;
    ColumnEnd = ColumnStart;
  }
}

// This function return the start position to the first column of the next line.
void Location::Cr( void )
{
  LineStart++;
  ColumnStart = 1;
  if ( POSITION(LineStart,ColumnStart) > POSITION(LineEnd,ColumnEnd) )
  {
    LineEnd = LineStart;
    ColumnEnd = ColumnStart;
  }
}

void Location::Select( int length )
{
  LineStart   = LineEnd;
  ColumnStart = ColumnEnd;
  ColumnEnd  += length - 1;
  if ( ColumnEnd >= LINE_WIDTH )
  {
    LineEnd++;
    ColumnEnd = (ColumnEnd - 1 ) % LINE_WIDTH + 1;
  }
}

void Location::Advance( void )
{
  LineStart = LineEnd;
  ColumnStart = ++ColumnEnd;
  if (ColumnStart == LINE_WIDTH)
  {
    ColumnStart = ColumnEnd = 1;
    LineStart = ++LineEnd;
  }
}

void Location::CrSelect( void )
{
  LineEnd++;
  ColumnEnd = 1;
}

// Checking the properties of a location Object :
// Line numbers and Column numbers must be greater than 0 and
// Start Mark must be come before the End Mark.

int Location::OK( void )
{
  if (  (LineStart   < 0) || (LineEnd   < 0) ||
        (ColumnStart < 0) || (ColumnEnd < 0) ||
// here we compute the number of characters from the begining
// assuming the line is LINE_WIDTH characters wide .
        ( POSITION( LineStart,ColumnStart ) > POSITION( LineEnd,ColumnEnd ) )
     )
  {
    error(" This Location object is not in a coherent state.");
    return -1;
  }
  else
    return 0;
}

//  Overloading semantic of + operator.
//  this return a Location Object that include both objects Location.
Location operator + ( const Location& FirstLocation  ,
                      const Location& SecondLocation )
{
  if (POSITION( FirstLocation .LineStart,FirstLocation .ColumnStart ) <
      POSITION( SecondLocation.LineEnd  ,SecondLocation.ColumnEnd   ) )
  {
    return Location( FirstLocation .LineStart  ,
                     FirstLocation .ColumnStart,
                     SecondLocation.LineEnd,
                     SecondLocation.ColumnEnd );
  }
  else
  {
    return Location( SecondLocation.LineStart  ,
                     SecondLocation.ColumnStart,
                     FirstLocation.LineEnd,
                     FirstLocation.ColumnEnd );
  }
}

// Well, this print out the contents of a Location object.

ostream& operator << ( ostream& cout, Location there )
{
  cout << '(' << there.LineStart << ':' << there.ColumnStart << ',';
  return cout << there.LineEnd << ':' << there.ColumnEnd << ')';
}

// General fonction that force the program to abort.
void error( char * TheString )
{
  cout << TheString;
  exit(-1);
}
